/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright (c) 2009, 2010, Oracle and/or its affiliates. All rights reserved.
 */

/*
 * ADR: Abstract Data Representation
 *
 * These are a set of interfaces for constructing arbitrarily complex,
 * self-describing data types, using a set of base types, arrays, and
 * structs as building blocks.
 *
 * The intent of these interfaces is not to build a higher-level set of
 * data structures on top of C, but to facilitate interchange with
 * other systems by precisely defining the structure of the data used
 * when communicating with them.  The former isn't precluded by the
 * latter, but it is against the latter that the utility of these
 * interfaces must always be measured.
 *
 * Reference counting
 *
 * Data objects are reference counted.  Since in most cases there is
 * only one owner of a piece of data at a time, the interfaces are
 * designed around such usage.  In particular, the interfaces which
 * assign values to structures and arrays assume that a caller that
 * needs to access the value after the assignment will obtain a
 * reference for itself.  Likewise, routines that obtain values from
 * complex types assume a caller that needs the value after it has
 * released its reference on the complex type will obtain a reference
 * for itself.  The effect of this design is that it is very rare that
 * ADR consumers will need to interact with the reference counting
 * interfaces.
 *
 * In addition to having interfaces optimized for the 1-reference
 * case, the implementation assumes multiple references will be rare.
 * If multiple references ever become the norm, the implementation
 * will need to be revised to avoid a bottleneck.
 *
 * Synchronization
 *
 * With the exception of the composite types, arrays and structures,
 * data objects are immutable.  Moreover, once an array or structure
 * has been made visible to other threads, they too are to be
 * considered immutable.  Synchronizing writes to data objects is
 * therefore unnecessary.
 *
 * Error handling
 *
 * To streamline error handling, many of the interfaces will:
 *   a) translate error input to error output, and
 *   b) dereference/free valid input when an error condition occurs
 *
 * In conjunction with data_verify()/data_purify(), these permit a
 * consumer to perform a long sequence of operations and check for
 * success only at the very end by examining the final product of
 * the operations.  Note that structure fields marked as optional
 * introduce an ambiguity: it is impossible to automatically
 * distinguish a field that is missing due to error from an field
 * that is intentionally missing.  Such error conditions will need
 * to be handled manually.
 */

#ifndef	_ADR_H
#define	_ADR_H

#include <sys/types.h>
#include <stdio.h>

#include <rad/adr_name.h>

#ifdef	__cplusplus
extern "C" {
#endif

typedef enum datatype {
	dt_void,
	dt_boolean,
	dt_int,
	dt_string,
	dt_array,
	dt_struct,
	dt_enum,
	dt_opaque,
	dt_long,
	dt_time,
	dt_ulong,
	dt_uint,
	dt_object,
	dt_password,
	dt_float,
	dt_double
} datatype_t;

typedef struct typefield {
	const char *tf_name;		/* field name */
	struct type *tf_type;		/* field type */
	boolean_t tf_optional;		/* field is optional */
} typefield_t;

/*
 * Enumeration values have both a name and value.  The values could be
 * represented positionally, but this lets us easily add support for
 * values with assigned integer values.
 */
typedef struct enumval {
	const char *ev_name;		/* value name */
	int ev_value;			/* value integer value */
	struct data *ev_data;		/* value's constant instantiation */
} enumval_t;

typedef struct type {
	datatype_t t_type;		/* basic type */
	const char *t_name;		/* type name */
	int t_size;			/* for structs, enums */
	union {
		struct type *t_array;
		typefield_t *t_fields;
		enumval_t *t_enum;
	} t_aux;
} type_t;


/* Lifetimes for externally allocated buffers */
typedef enum lifetime {
	lt_copy,	/* Make a copy (argument only; becomes lt_free) */
	lt_free,	/* Free when done */
	lt_const,	/* Will never change, shouldn't be freed */
	lt_ref		/* Reference counted (currently unused) */
} lifetime_t;

typedef struct data {
	type_t *d_type;
	unsigned int d_size;		/* for complex types */
	unsigned int d_rsize;
	int d_refs;			/* 0 = constant */
	lifetime_t d_lifetime;		/* lifetime of string/opaque data */
	union {
		boolean_t boolean;
		int integer;		/* also used for enumerations */
		unsigned int uinteger;
		long long longint;
		unsigned long long ulongint;
		const char *string;
		struct data **array;
		void *opaque;
		adr_name_t *object;
		float afloat;
		double adouble;
	} d_data;
} data_t;

typedef enum stability {
	s_undefined,
	s_private,
	s_uncommitted,
	s_committed
} stability_t;

const char *strtype(datatype_t);
data_t *data_ref(data_t *);
data_t *data_ref_n(data_t *, unsigned int);
void data_free(data_t *);
void data_print(FILE *, data_t *);
void type_print(FILE *, type_t *);
void type_free(type_t *);
boolean_t data_verify(data_t *, type_t *, boolean_t);
data_t *data_purify(data_t *);

/*
 * Data constructors
 */
data_t *data_new_string(const char *, lifetime_t);
data_t *data_new_fstring(const char *, ...);
data_t *data_new_nstring(const char *, int);
data_t *data_new_boolean(boolean_t);
data_t *data_new_integer(int);
data_t *data_new_long(long long);
data_t *data_new_uinteger(unsigned int);
data_t *data_new_ulong(unsigned long long);
data_t *data_new_time(time_t);
data_t *data_new_struct(type_t *);
data_t *data_new_array(type_t *, int);
data_t *data_new_enum(type_t *, int);
data_t *data_new_enum_byname(type_t *, const char *);
data_t *data_new_opaque(void *, int, lifetime_t);
data_t *data_new_object(adr_name_t *);
data_t *data_new_enum_index(type_t *, int);	/* Private interface */
data_t *data_new_password(const char *);
data_t *data_new_float(float);
data_t *data_new_double(double);

/*
 * Data access routines/macros
 */

#ifdef NDEBUG
#define	data_get(data, type, field)	(data->d_data.field)
#else
/* We use a helper function to avoid double-evaluating 'data'. */
#define	data_get(data, type, field)	\
	(data_get_internal(data, type)->d_data.field)
#endif

#define	data_to_string(data)	data_get(data, dt_string, string)
#define	data_to_integer(data)	data_get(data, dt_int, integer)
#define	data_to_uinteger(data)	data_get(data, dt_uint, uinteger)
#define	data_to_longint(data)	data_get(data, dt_long, longint)
#define	data_to_ulongint(data)	data_get(data, dt_ulong, ulongint)
#define	data_to_boolean(data)	data_get(data, dt_boolean, boolean)
#define	data_to_object(data)	data_get(data, dt_object, object)
#define	data_to_password(data)  data_get(data, dt_password, string);
#define	data_to_float(data)	data_get(data, dt_float, afloat);
#define	data_to_double(data)	data_get(data, dt_double, adouble);

/*
 * Type routines
 */
datatype_t data_basetype(data_t *);
type_t *data_type(data_t *);
type_t *array_type(data_t *);
data_t *data_get_internal(data_t *, datatype_t);

/*
 * Enum routines
 */
const char *enum_tostring(data_t *);
int enum_tovalue(data_t *);

/*
 * Array routines
 */
void array_set(data_t *, int, data_t *);
int array_vset(data_t *, int, data_t *);
int array_add(data_t *, data_t *);
void array_remove(data_t *, int);
data_t *array_combine(data_t *, data_t *);
int array_search(data_t *, const char *, const char * const *);
int array_nsearch(data_t *, const char *, int, const char * const *);

#define	array_get(data, index)	data_get(data, dt_array, array[index])
#define	array_size(data)	((data)->d_rsize)

/*
 * Structure routines
 */
void struct_set(data_t *, const char *, data_t *);
void struct_nset(data_t *, const char *, int, data_t *);
void struct_merge(data_t *, data_t *);
data_t *struct_get(data_t *, const char *);
typefield_t *type_struct_get(type_t *, const char *);
typefield_t *type_struct_nget(type_t *, const char *, int);

/*
 * Utility functions
 */
int opaque_cmp(data_t *, data_t *);
int string_cmp(data_t *, data_t *);
data_t *rad_split(const char *, const char *, boolean_t);
data_t *rad_strarray(const char **, int, lifetime_t);

/*
 * Pre-fab basic type definitions
 */
extern type_t t_void;
extern type_t t_string;
extern type_t t_integer;
extern type_t t_uinteger;
extern type_t t_long;
extern type_t t_ulong;
extern type_t t_time;
extern type_t t_object;
extern type_t t_boolean;
extern type_t t_opaque;
extern type_t t_password;
extern type_t t_float;
extern type_t t_double;

extern type_t t_array_string;
extern type_t t_array_integer;
extern type_t t_array_uinteger;
extern type_t t_array_long;
extern type_t t_array_ulong;
extern type_t t_array_time;
extern type_t t_array_object;
extern type_t t_array_boolean;
extern type_t t_array_opaque;
extern type_t t_array_password;
extern type_t t_array_float;
extern type_t t_array_double;

#ifdef	__cplusplus
}
#endif

#endif	/* _ADR_H */
