/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 */

/*
 * ADR Object Model
 *
 * These are a set of interfaces for defining a basic object model,
 * built using ADR types.
 */

#ifndef	_ADR_OBJECT_H
#define	_ADR_OBJECT_H

#include <sys/types.h>
#include <stdio.h>
#include <rad/adr.h>

#ifdef	__cplusplus
extern "C" {
#endif

/* A method parameter */
typedef struct adr_parameter {
	const char *ap_name;		/* argument's name */
	type_t *ap_type;		/* argument's type */
	boolean_t ap_optional;		/* can this argument be "null"? */
} adr_parameter_t;

/* An object method */
typedef struct adr_method {
	const char *am_name;		/* method's name */
	stability_t am_stability;	/* method stability classification */
	int am_nargs;			/* number of arguments */
	type_t *am_result;		/* result type */
	type_t *am_error;		/* error type */
	boolean_t am_optional;		/* can the result be "null"? */
	adr_parameter_t *am_args;	/* method's arguments */
} adr_method_t;

/* An object attribute */
typedef struct adr_attribute {
	const char *aa_name;		/* attribute's name */
	stability_t aa_stability;	/* attribute stability classification */
	type_t *aa_type;		/* attribute's type */
	type_t *aa_read_error;		/* attribute's read error type */
	type_t *aa_write_error;		/* attribute's write error type */
	boolean_t aa_optional;		/* can this attribute be "null"? */
	boolean_t aa_readable;		/* can this attribute be read */
	boolean_t aa_writeable;		/* can this attribute be written */
} adr_attribute_t;

/* An object event */
typedef struct adr_event {
	const char *ae_name;		/* event's name */
	stability_t ae_stability;	/* event's stability classification */
	type_t *ae_type;		/* event's type */
} adr_event_t;

/* An object, consisting of attributes and methods */
typedef struct adr_object {
	const char *ao_name;		/* name of type */
	int ao_major;			/* major version */
	int ao_minor;			/* minor version */
	int ao_nattributes;		/* # attributes */
	int ao_nmethods;		/* # methods */
	int ao_nevents;			/* # events */
	adr_attribute_t **ao_attributes; /* attributes */
	adr_method_t **ao_methods;	/* methods */
	adr_event_t *ao_events;		/* events */
} adr_object_t;

void adr_free(adr_object_t *);

int adr_lookup_attr_index(adr_object_t *, const char *);
adr_attribute_t *adr_lookup_attr(adr_object_t *, const char *);
adr_attribute_t *adr_get_attr(adr_object_t *, int);

int adr_lookup_method_index(adr_object_t *, const char *);
adr_method_t *adr_lookup_method(adr_object_t *, const char *);
adr_method_t *adr_get_method(adr_object_t *, int);

int adr_lookup_event_index(adr_object_t *, const char *);
adr_event_t *adr_lookup_event(adr_object_t *, const char *);
adr_event_t *adr_get_event(adr_object_t *, int);

#ifdef	__cplusplus
}
#endif

#endif	/* _ADR_OBJECT_H */
