#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
#

import xdrlib
import threading
import rad.recordmarking as recordmarking

class Op(object):
    INVOKE = 0
    GETATTR = 1
    SETATTR = 2
    DESCRIBE = 3
    LIST = 4
    SUB = 5
    UNSUB = 6

class Type(object):
    VOID = 0
    BOOLEAN = 1
    INTEGER = 2
    STRING = 3
    ARRAY = 4
    COMPOSITE = 5
    ENUM = 6
    OPAQUE = 7
    LONG = 8
    DATE = 9
    ULONG = 10
    UINT = 11
    OBJECT = 12
    PASSWORD = 13
    FLOAT = 14
    DOUBLE = 15

class Error(object):
    OK = 0
    OBJECT = 1
    NOMEM = 2
    NOTFOUND = 3
    PRIV = 4
    SYSTEM = 5
    EXISTS = 6
    MISMATCH = 7
    ILLEGAL = 8

class RequestError(Exception):
    def __init__(self, unpacker):
	self._error = unpacker.unpack_int()
	self._source = unpacker.unpack_string()
	self._message = unpacker.unpack_string()
	self._payload = unpacker.unpack_opaque()

    def __str__(self):
	return 'Request error: ' + self.get_message()

    def get_message(self):
	return '%s (%d)' % (errors[self._error][0], self._error)

class ContainerError(Exception):
    def __init__(self, msg, ex, error = None):
    	self._msg = msg
    	self._error = ex
	if error:
		self._payload = error.read(xdrlib.Unpacker(ex._payload), True)
	else:
		self._payload = None

    def get_payload(self):
    	return self._payload

    def get_message(self):
    	return self._error.get_message()

    def __str__(self):
    	if self._payload:
	    return self._payload.__str__()
	else:
	    return errors[self._error._error][0]

class ObjectError(ContainerError):
    pass

class NomemError(ContainerError):
    pass

class NotFoundError(ContainerError):
    pass

class PrivError(ContainerError):
    pass

class SystemError(ContainerError):
    pass

class ExistsError(ContainerError):
    pass

class MismatchError(ContainerError):
    pass

class IllegalError(ContainerError):
    pass

errors = {
    Error.OK : [ 'no error', None ],
    Error.OBJECT : [ 'object-specific error', ObjectError ],
    Error.NOMEM : [ 'out of memory', NomemError ],
    Error.NOTFOUND : [ 'not found', NotFoundError ],
    Error.PRIV : [ 'insufficient privileges', PrivError ],
    Error.SYSTEM : [ 'system error', SystemError ],
    Error.EXISTS : [ 'already exists', ExistsError ],
    Error.MISMATCH : [ 'type mismatch', MismatchError ],
    Error.ILLEGAL : [ 'illegal access', IllegalError ]
}

def raise_error(ex, msg = None):
    raise errors[ex._error][1](msg, ex, None)

# Should only be used when there is a mismatch, but is currently used
# as a catch-all.  Fix.

class InvalidFeatureError(Exception):
    def __init__(self, name, msg = None):
    	self._name = name
	self._msg = msg

    def __str__(self):
	return self.get_message()

    def get_name(self):
    	return self._name

    def get_message(self):
    	if self._msg: return self._msg
	return 'Feature not found: ' + self._name

class Name(object):
    def __init__(self, domain, kvs):
    	if domain:
	    self._domain = domain
	else:
	    self._domain = ""
	self._keys = dict(kvs)
	self._sortedkeys = sorted(kvs, key = lambda kv: kv[0])

    @staticmethod
    def quote(value):
	return value.replace("\\", "\\S").replace("=",
	     "\\E").replace(",", "\\C")

    @staticmethod
    def unquote(value):
	return value.replace("\\E", "=").replace("\\C",
	     ",").replace("\\S", "\\")

    @staticmethod
    def unpack(unpacker):
	pass

    @staticmethod
    def parse(namestr):
    	parts = namestr.split(":", 1)
	if len(parts) < 2:
		raise Exception("Bad name format: %s" % namestr)

	domain = parts[0]
	if len(parts[1]) == 0:
		return Name(domain, [])

	kvs = []
	kvstr = parts[1].split(",")
	for i in kvstr:
		kv = i.split("=")
		if len(kv) != 2:
			raise Exception("Bad key/value: %s" % i)
		kvs.append([Name.unquote(kv[0]), Name.unquote(kv[1])])

	return Name(domain, kvs)

    def match(self, name):
        if len(self._domain) > 0 and self._domain != name._domain:
	    return False
	for key, value in self._sortedkeys:
	    if not key in name._keys:
		return False
	    if name._keys[key] != value:
		return False
	return True

    def get_domain(self):
	return self._domain

    def get_value(self, key):
	return self._keys[key]

    def num_keys(self):
	return len(self._sortedkeys)

    def __cmp__(self, other):
    	if not other:
	    return 1

    	if self._domain < other._domain:
	    return -1
	if self._domain > other._domain:
	    return 1
	
	scount = len(self._sortedkeys)
	ocount = len(other._sortedkeys)

	if scount < ocount:
	    return -1
	if scount > ocount:
	    return 1

	for i in range(scount):
	    if self._sortedkeys[i][0] < other._sortedkeys[i][0]:
		return -1
	    if self._sortedkeys[i][0] > other._sortedkeys[i][0]:
		return 1
	    if self._sortedkeys[i][1] < other._sortedkeys[i][1]:
		return -1
	    if self._sortedkeys[i][1] > other._sortedkeys[i][1]:
		return 1

	return 0

    def __str__(self):
        keystr = ",".join([ "%s=%s" % (Name.quote(i[0]), Name.quote(i[1]))
	    for i in self._sortedkeys])
	return "%s:%s" % (self._domain, keystr)

    def __repr__(self):
    	return "Name.parse(\"%s\")" % str(self)

class RadField(object):
    def __init__(self, unpacker, obj):
	self._name = unpacker.unpack_string()
	self._optional = unpacker.unpack_bool()
	self._type = RadType(unpacker, obj)

class RadEnumVal(object):
    def __init__(self, unpacker):
	self._name = unpacker.unpack_string()
	self._value = unpacker.unpack_uint()

#
# Superclass for synthetic structure objects
#
class RadData(object):
    def __init__(self, type):
	self._type = type

    def __repr__(self):
    	return "%s(%s)" % (self._type._name, ", ".join(map(
	    lambda f: "%s = %r" % (f._name, self.__dict__[f._name]),
	    self._type._fields.values())))

class RadType(object):
    def __init__(self, unpacker, obj):
	self._type = unpacker.unpack_uint();
	if self._type == Type.ARRAY:
	    self._atype = RadType(unpacker, obj)
	elif self._type == Type.COMPOSITE:
	    self._name = unpacker.unpack_string();
	    self._pfields = unpacker.unpack_array(
	        lambda: RadField(unpacker, obj))
	    self._fields = dict([(i._name, i) for i in self._pfields])
	    if obj: obj._types[self._name] = self;
	elif self._type == Type.ENUM:
	    self._name = unpacker.unpack_string();
	    self._values = dict([(i._name, i) for i in
	        unpacker.unpack_array(lambda: RadEnumVal(unpacker))])
	    if obj: obj._types[self._name] = self;

    def make(self, args):
	if self._type != Type.COMPOSITE:
	    raise Exeception("make() only invokable on composite types")
	o = RadData(self);
	for i in self._pfields:
	    if i._name in args:
		o.__dict__[i._name] = args[i._name]
	    else:
		if not i._optional:
		    raise Exception("Field '%s' of '%s' must be specified" %
			(i._name, self._name))
		o.__dict__[i._name] = None
	return o

    def makev(self, *args):
	if self._type == Type.COMPOSITE:
	    o = RadData(self);
	    n = 0
	    for i in self._pfields:
		if len(args) > n:
		    o.__dict__[i._name] = args[n]
	        else:
		    if not i._optional:
		        raise Exception("Field '%s' of '%s' must be specified" %
			    (i._name, self._name))
		    o.__dict__[i._name] = None
	        n = n + 1
	    return o
	if self._type == Type.ENUM:
	    if len(args) != 1:
	        raise Exception("Type '%s' is an enumeration" % self._name)
	    for vals in self._values:
	    	if args[0] == self._values:
		    return args[0];
	    raise Exception("Enumeration '%s' doesn't contain value '%s'." %
	        (self._name, args[0]))
        if len(args) != 1:
	    raise Exception("Type '%s' has one value" % self._name)
	return args[0]

    def read(self, unpacker, optional=False):
	if optional and not unpacker.unpack_bool(): return None
	if self._type == Type.VOID:
	    pass
	elif self._type == Type.BOOLEAN:
	    return unpacker.unpack_bool()
	elif self._type == Type.INTEGER:
	    return unpacker.unpack_int()
	elif self._type == Type.STRING:
	    return unpacker.unpack_string()
	elif self._type == Type.ARRAY:
	    return unpacker.unpack_array(lambda: self._atype.read(unpacker))
	elif self._type == Type.COMPOSITE:
	    o = RadData(self)
	    for i in self._pfields:
		# order isn't important, yet
		name = unpacker.unpack_string()
		o.__dict__[name] = self._fields[name]._type.read(unpacker, True)
	    return o
	elif self._type == Type.ENUM:
	    return unpacker.unpack_string();
	elif self._type == Type.OPAQUE:
	    return unpacker.unpack_opaque()
	elif self._type == Type.LONG:
	    return unpacker.unpack_hyper()
	elif self._type == Type.DATE:
	    return unpacker.unpack_string()
	elif self._type == Type.ULONG:
	    return unpacker.unpack_uhyper()
	elif self._type == Type.UINT:
	    return unpacker.unpack_uint()
	elif self._type == Type.OBJECT:
	    return Name.parse(unpacker.unpack_string())
	elif self._type == Type.PASSWORD:
	    return unpacker.unpack_string()
	elif self._type == Type.FLOAT:
	    return unpacker.unpack_float()
	elif self._type == Type.DOUBLE:
	    return unpacker.unpack_double()

    def write(self, packer, data, optional=False):
	if optional:
	    isdata = data != None
	    packer.pack_bool(isdata)
	    if not isdata: return
	if self._type == Type.VOID:
	    pass
	elif self._type == Type.BOOLEAN:
	    return packer.pack_bool(data)
	elif self._type == Type.INTEGER:
	    return packer.pack_int(data)
	elif self._type == Type.STRING:
	    return packer.pack_string(data)
	elif self._type == Type.ARRAY:
	    return packer.pack_array(data,
		lambda x: self._atype.write(packer, x))
	elif self._type == Type.COMPOSITE:
	    for i in self._pfields:
		# order isn't important, yet
		packer.pack_string(i._name)
		if i._name in data.__dict__:
		    i._type.write(packer, data.__dict__[i._name], True)
		else:
		    i._type.write(packer, None, True)
	elif self._type == Type.ENUM:
	    return packer.pack_string(data);
	elif self._type == Type.OPAQUE:
	    return packer.pack_opaque(data)
	elif self._type == Type.LONG:
	    return packer.pack_hyper(data)
	elif self._type == Type.DATE:
	    return packer.pack_string(data)
	elif self._type == Type.ULONG:
	    return packer.pack_uhyper(data)
	elif self._type == Type.UINT:
	    return packer.pack_uint(data)
	elif self._type == Type.OBJECT:
	    return packer.pack_string(str(data))
	elif self._type == Type.PASSWORD:
	    return packer.pack_string(data)
	elif self._type == Type.FLOAT:
	    return packer.pack_float(data);
	elif self._type == Type.DOUBLE:
	    return packer.pack_double(data);

class RadAttribute(object):
    def __init__(self, unpacker, obj):
	self._name = unpacker.unpack_string()
	self._stability = unpacker.unpack_enum()
	self._readable = unpacker.unpack_bool()
	self._writable = unpacker.unpack_bool()
	self._optional = unpacker.unpack_bool()
	self._type = RadType(unpacker, obj)
	if unpacker.unpack_bool():
		self._read_error = RadType(unpacker, obj)
	else:
		self._read_error = None
	if unpacker.unpack_bool():
		self._write_error = RadType(unpacker, obj)
	else:
		self._write_error = None

    def read(self, instance):
	if not self._readable:
	    raise InvalidFeatureError(self._name,
	        'Attribute %s of %s isn\'t readable' %
		(self._name, instance._name))

	p = xdrlib.Packer()
	p.pack_hyper(instance._id)
	p.pack_string(self._name)
	conn = instance._connection
	ser = conn.make_request(Op.GETATTR, p.get_buffer())

	try:
	    return self._type.read(conn.read_response(ser, None), True)
	except RequestError, ex:
	    if ex._error == 1:
	    	raise ObjectError("Object-specific error", ex, self._read_error)
	    raise_error(ex, 'Error getting %s from %s: %d' %
		(self._name, instance._name, ex._error))

    def write(self, instance, value):
	if not self._writable:
	    raise InvalidFeatureError(self._name,
	        'Attribute %s of %s isn\'t writable' %
		(self._name, instance._name))

	p = xdrlib.Packer()
	p.pack_hyper(instance._id)
	p.pack_string(self._name)
	data = xdrlib.Packer()
	self._type.write(data, value, True)
	p.pack_opaque(data.get_buffer())
	conn = instance._connection
	ser = conn.make_request(Op.SETATTR, p.get_buffer())

	try:
	    conn.read_response(ser, None)
	except RequestError, ex:
	    if ex._error == 1:
	    	raise ObjectError("Object-specific error", ex, self._write_error)
	    raise_error(ex, 'Error setting %s on %s: %d' %
		(self._name, instance._name, ex._error))


class RadArg(object):
    def __init__(self, unpacker, obj):
	self._name = unpacker.unpack_string()
	self._optional = unpacker.unpack_bool()
	self._type = RadType(unpacker, obj)

class RadMethod(object):
    def __init__(self, object, unpacker):
	self._object = object
	self._name = unpacker.unpack_string()
	self._stability = unpacker.unpack_enum()
	self._optional = unpacker.unpack_bool()
	self._type = RadType(unpacker, object)
	if unpacker.unpack_bool():
		self._error = RadType(unpacker, object)
	else:
		self._error = None
	self._args = unpacker.unpack_array(lambda : RadArg(unpacker, object))

    def call(self, *args):
	if len(self._args) != len(args):
	    raise InvalidFeatureError(self._name,
		"%s called with invalid # of args (received %d, expected %d)" %
		(self._name, len(args), len(self._args)))

	p = xdrlib.Packer()
	p.pack_hyper(self._object._id)
	p.pack_string(self._name)
	p.pack_uint(len(args))
	for i in range(len(args)):
	    pp = xdrlib.Packer()
	    self._args[i]._type.write(pp, args[i], True)
	    p.pack_opaque(pp.get_buffer())
	conn = self._object._connection
	ser = conn.make_request(Op.INVOKE, p.get_buffer())

	try:
	    return self._type.read(conn.read_response(ser, None), True)
	except RequestError, ex:
	    if ex._error == 1:
	    	raise ObjectError("Object-specific error", ex, self._error)
	    raise_error(ex, 'Error invoking %s on %s: %d' %
		(self._name, self._object._name, ex._error))

class RadEventType(object):
    def __init__(self, unpacker, obj):
	self._name = unpacker.unpack_string()
	self._stability = unpacker.unpack_enum()
	self._type = RadType(unpacker, obj)

class RadObject(object):
    def __init__(self, connection, name, unpacker):
	self._connection = connection
	self._name = name
        self._types = {}
	self._id = unpacker.unpack_hyper()
	self._major = unpacker.unpack_int()
	self._minor = unpacker.unpack_int()
	self._attributes = dict([[i._name, i] for i in
	    unpacker.unpack_array(lambda : RadAttribute(unpacker, self))])
	self._methods = dict([[i._name, i] for i in
	    unpacker.unpack_array(lambda : RadMethod(self, unpacker))])
	self._events = dict([[i._name, i] for i in
	    unpacker.unpack_array(lambda : RadEventType(unpacker, self))])

    def get_type(self, name):
    	return self._types[name]

    def read(self, attr):
	try:
    	    return self._attributes[attr].read(self)
	except KeyError:
	    raise InvalidFeatureError(self._name,
	        "%s has no attribute %s" % (self._name, attr))

    def write(self, attr, value):
	try:
	    self._attributes[attr].write(self, value)
	except KeyError:
	    raise InvalidFeatureError(self._name,
	        "%s has no attribute %s" % (self._name, attr))

    def call(self, meth, *args):
    	return self._methods[meth].call(*args)

    def __repr__(self):
        return "< Remote RAD object: %s >" % self._name

class RadEvent(object):
    def __init__(self, unpacker):
	self._source = unpacker.unpack_hyper()
	self._sequence = unpacker.unpack_hyper()
	self._tstamp = unpacker.unpack_hyper()
	self._event = unpacker.unpack_string()
	self._payload = unpacker.unpack_opaque()

class RadReader(threading.Thread):
    def __init__(self, connection):
	threading.Thread.__init__(self)
	self.daemon = True
    	self._connection = connection

    def run(self):
    	try:
		while True:
		    msg, serial = self._connection.read_message()
		    self._connection.post_message(msg, serial)
	except Exception, e:
		self._connection.set_error(e)
    	
class RadConnection(object):
    def __init__(self, socket, threaded=True):
    	self._wlock = threading.Lock()
    	self._rlock = threading.Lock()
    	self._tlock = threading.Lock()
    	self._econd = threading.Condition(self._rlock)
    	self._rcond = threading.Condition(self._rlock)
	self._closed = None
	self._responses = { }
	self._events = []
	self._errors = [ None, None ]	# no error and object-specific error
	self._rs = recordmarking.RecordMarkingSocket(socket)
	self._reader = None

	self._rs.skip_record()
	data = self._rs.read_record()
	vers = xdrlib.Unpacker(data)
	name = vers.unpack_string()
	min = vers.unpack_int()
	max = vers.unpack_int()

	response = xdrlib.Packer()
	response.pack_string(name)
	response.pack_int(1)
	self._rs.write(response.get_buffer(), True)

	self._rs.skip_record()
	data = self._rs.read_record()
	errors = xdrlib.Unpacker(data)
	self._errors.extend(errors.unpack_array(lambda: RadType(errors, None)))

	self._serial = 1

	if threaded:
	    self.start_threads()

    def start_threads(self):
    	with self._tlock:
	    if not self._reader:
		self._reader = RadReader(self)
		self._reader.start()

    def set_error(self, e):
    	with self._rlock:
	    self._closed = e
	    self._econd.notifyAll()
	    self._rcond.notifyAll()

    def read_message(self):
        self._rs.skip_record()
	up = xdrlib.Unpacker(self._rs.read_record())
	return up, up.unpack_uint()

    def post_message(self, up, serial):
	with self._rlock:
	    if serial == 0:
		self._events.append(RadEvent(up))
		self._econd.notifyAll()
	    else:
		self._responses[serial] = up
		self._rcond.notifyAll()

    def unpack_message(self, up):
	success = up.unpack_bool()
	if success:
	    payload = up.unpack_opaque()
	    up.reset(payload)
	    return up
	else:
	    raise RequestError(up)

    def read_response(self, reqserial, etype):
        with self._tlock:
	    if not self._reader:
	    	up, serial = self.read_message()
		if serial != reqserial:
		    raise Exception('Bad response (serial %d, expected %d)' %
			(serial, reqserial))
	    	return self.unpack_message(up)

	with self._rlock:
	    while not reqserial in self._responses:
	    	if self._closed:
			raise IOError, self._closed
	    	self._rcond.wait()
	    up = self._responses.pop(reqserial)
	    return self.unpack_message(up)

    def make_request(self, op, buffer):
    	with self._wlock:
	    serial = self._serial
	    p = xdrlib.Packer()
	    p.pack_uint(serial)
	    p.pack_enum(op)
	    p.pack_opaque(buffer)
	    self._serial += 1
	    self._rs.write(p.get_buffer(), True)
	    return serial

    def list_objects(self, pattern = None):
	p = xdrlib.Packer()
	if pattern:
		p.pack_string(str(pattern))
	else:
		p.pack_string('')
	ser = self.make_request(Op.LIST, p.get_buffer())
	up = self.read_response(ser, None)
	return [ Name.parse(i) for i in up.unpack_array(up.unpack_string) ]

    def get_object(self, name):
	p = xdrlib.Packer()
	p.pack_string(str(name))
	ser = self.make_request(Op.DESCRIBE, p.get_buffer())
	try:
		up = self.read_response(ser, None)
	except RequestError, ex:
		raise_error(ex)
	return RadObject(self, name, up)

    def get_object_s(self, domain, keys):
	return self.get_object(Name(domain, keys))

    def subscribe(self, object, event):
	p = xdrlib.Packer()
	p.pack_hyper(object._id)
	p.pack_string(event)
	ser = self.make_request(Op.SUB, p.get_buffer())
	try:
		self.read_response(ser, None)
	except RequestError, ex:
		raise_error(ex)

    def unsubscribe(self, object, event):
	p = xdrlib.Packer()
	p.pack_hyper(object._id)
	p.pack_string(event)
	ser = self.make_request(Op.UNSUB, p.get_buffer())
	try:
		self.read_response(ser, None)
	except RequestError, ex:
		raise_error(ex)

    def read_event(self):
    	with self._rlock:
	    while len(self._events) == 0:
	    	if self._closed:
			raise IOError, self._closed
	    	self._econd.wait()
	    return events.pop(0)
