#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
#

import socket
import ssl
import getpass
import subprocess
import os
import rad.client as rad
import rad.adaptor as adapt

#
# Make pipes quack enough like a socket to satisfy RecordMarkingSocket
#
class SocketLike(object):
	def __init__(self, inf, outf):
		self._in = inf
		self._out = outf

	def recv(self, n):
		#
		# If the file is unbuffered, it reads one byte at a time.
		# If the file is buffered, it waits until all n bytes are
		# read before returning.
		# Neither is acceptable, so we use os.read.
		#
		return os.read(self._out.fileno(), n)

	def sendall(self, buf):
		r = self._in.write(buf)
		self._in.flush()
		return r

#
# Connect to a privately spawned copy of rad
#
def connect_private():
	p = subprocess.Popen(["/usr/lib/rad/rad",
	    "-m", "/usr/lib/rad/transport",
	    "-m", "/usr/lib/rad/protocol",
	    "-m", "/usr/lib/rad/module",
	    "-t", "stdin:exit"],
	    0, None, subprocess.PIPE, subprocess.PIPE)

	return rad.RadConnection(SocketLike(p.stdin, p.stdout))
	
#
# Connect to a remote rad daemon using SSH
# The remote system must have netcat (/usr/bin/nc) installed
#
def connect_ssh(host, user = None):
	if user:
		host = "%s@%s" % (user, host)
	p = subprocess.Popen(["/usr/bin/ssh", "-q", "-e", "none",
	    "-o", "BatchMode yes", host, "/usr/lib/rad/radpipe" ],
	    0, None, subprocess.PIPE, subprocess.PIPE)
	return rad.RadConnection(SocketLike(p.stdin, p.stdout))

#
# Connect to the local rad daemon via the standard unix domain socket
#
def connect_local():
	s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM);
	s.connect("/var/run/radsocket")
	return rad.RadConnection(s)

#
# Connect to a remote rad daemon at the standard port
#
def connect_ssl(host):
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM);
	s.connect((host, 12302))
	s = ssl.wrap_socket(s)
	return rad.RadConnection(s)

# 
# Fetches and caches a handle to the authentication object.
# Supplies some convenience routines built on top of it.
#
class RadAuth(object):
	AUTH_NAME = rad.Name("org.opensolaris.os.rad",
	    [("type", "authentication")])

	def __init__(self, rc):
		self._auth = adapt.Adaptor(rc.get_object(self.AUTH_NAME))

	def handle(self):
		return self._auth

	#
	# Perform a terminal-interactive PAM login
	# Currently not recommended due to bug in Python's getpass.getpass()
	#
	def pam_login(self, user = None):
		if not user:
			user = raw_input("Username: ")
		blk = self._auth.login("C", user);
		while True:
			resp = []
			if blk.type == "success":
				break
			if blk.type == "error":
				raise Exception("Authentication Failed")
			for m in blk.messages:
				if m.style == "prompt_echo_off":
					r = getpass.getpass(m.message)
					resp.append(r)
				elif m.style == "prompt_echo_on":
					r = raw_input(m.message)
					resp.append(r)
				elif m.style == "text_info":
					print m.message
				elif m.style == "error_msg":
					print "ERROR: %s" % m.message
			blk = self._auth.submit(resp)

		self._auth.complete()
